import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.tabs.assignment_tab.assignment_tree import AssignmentTree
from qcanvas.ui.course_viewer.tabs.content_tab import ContentTab
from qcanvas.ui.course_viewer.tabs.util import date_strftime_format
from qcanvas.util.basic_fonts import bold_label
from qcanvas.util.layouts import grid_layout

_logger = logging.getLogger(__name__)


class AssignmentTab(ContentTab):
    @staticmethod
    def create_from_receipt(
        *,
        course: db.Course,
        sync_receipt: Optional[SyncReceipt],
        downloader: ResourceManager,
    ) -> "AssignmentTab":
        return AssignmentTab(
            course=course, sync_receipt=sync_receipt, downloader=downloader
        )

    def __init__(
        self,
        *,
        course: db.Course,
        sync_receipt: Optional[SyncReceipt],
        downloader: ResourceManager,
    ):
        super().__init__(
            explorer=AssignmentTree.create_from_receipt(
                course, sync_receipt=sync_receipt
            ),
            title_placeholder_text="No assignment selected",
            downloader=downloader,
        )

        self._due_date_label = QLabel("")
        self._score_label = QLabel("")

        self.enable_info_grid()

    def setup_info_grid(self) -> QGridLayout:
        grid = grid_layout(
            [
                [
                    bold_label("Due:"),
                    self._due_date_label,
                ],
                [
                    bold_label("Score:"),
                    self._score_label,
                ],
            ]
        )

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 1)

        return grid

    def update_info_grid(self, assignment: db.Assignment) -> None:
        if assignment.due_date is not None:
            due_text = assignment.due_date.strftime(date_strftime_format)
        else:
            due_text = "?"

        self._due_date_label.setText(due_text)

        self._score_label.setText(
            f"{assignment.mark or '?'}/{assignment.max_mark_possible or '?'}"
        )

    # def __init__(
    #     self,
    #     course: db.Course,
    #     resource_manager: ResourceManager,
    #     *,
    #     initial_sync_receipt: Optional[SyncReceipt] = None,
    # ):
    #     super().__init__()
    #     self._assignment_tree = AssignmentTree(
    #         course, sync_receipt=initial_sync_receipt
    #     )
    #     self._placeholder_assignment_title = "No assignment selected"
    #     self._assignment_label = QLabel(self._placeholder_assignment_title)
    #     self._assignment_label.setFont(bold_font)
    #     self._due_date_label = QLabel("")
    #     self._score_label = QLabel("")
    #     self._info_grid = self._setup_info_grid()
    #     make_truncatable(self._assignment_label)
    #     self._viewer = ResourceRichBrowser(resource_manager)
    #     self.setLayout(self._setup_layout())
    #     self._assignment_tree.assignment_selected.connect(self._assignment_selected)
    #     self._assignment_tree.reexpand()
    #
    # def _setup_info_grid(self) -> QWidget:
    #     layout = grid_layout(
    #         [
    #             [
    #                 _bold_label("Due:"),
    #                 self._due_date_label,
    #             ],
    #             [
    #                 _bold_label("Score:"),
    #                 self._score_label,
    #             ],
    #         ]
    #     )
    #     layout.setColumnStretch(0, 0)
    #     layout.setColumnStretch(1, 1)
    #
    #     widget = QWidget()
    #     widget.setLayout(layout)
    #
    #     return widget
    #
    # def _setup_layout(self) -> QHBoxLayout:
    #     h_box = QHBoxLayout()
    #     h_box.addWidget(self._assignment_tree, 1)
    #     h_box.addLayout(
    #         layout(QVBoxLayout, self._assignment_label, self._info_grid, self._viewer),
    #         2,
    #     )
    #
    #     return h_box
    #
    # def reload(self, course: db.Course, *, sync_receipt: Optional[SyncReceipt]) -> None:
    #     self._assignment_tree.reload(course, sync_receipt=sync_receipt)
    #
    # @Slot()
    # def _assignment_selected(self, assignment: db.Assignment) -> None:
    #     if assignment is not None:
    #         _logger.debug(
    #             "Show assignment %s (id='%s')", assignment.name, assignment.id
    #         )
    #         self._viewer.show_content(assignment)
    #         self._update_info(assignment)
    #     else:
    #         self._assignment_label.setText(self._placeholder_assignment_title)
    #         self._show_blank()
    #
    # def _update_info(self, assignment: db.Assignment) -> None:
    #     _logger.debug(assignment.course.name)
    #     _logger.debug("Due %s", assignment.due_date)
    #     _logger.debug("Score %s/%s", assignment.mark, assignment.max_mark_possible)
    #     self._assignment_label.setText(assignment.name)
    #
    #     if assignment.due_date is not None:
    #         self._due_date_label.setText(
    #             assignment.due_date.strftime("%A %Y-%m-%d %H:%M:%S")
    #         )
    #     else:
    #         self._due_date_label.setText("?")
    #     self._score_label.setText(
    #         f"{assignment.mark or '?'}/{assignment.max_mark_possible or '?'}"
    #     )
    #
    # def _show_blank(self) -> None:
    #     self._viewer.show_blank()
    #     self._info_grid.hide()



import logging
from typing import *

from qcanvas_api_clients.canvas import CanvasClientConfig
from qcanvas_api_clients.panopto import PanoptoClientConfig

from qcanvas.util import paths
from qcanvas.util.settings._mapped_setting import BoolSetting, MappedSetting

_logger = logging.getLogger(__name__)


class _ClientSettings:
    settings = paths.client_settings()
    canvas_url: MappedSetting[Optional[str]] = MappedSetting(default=None)
    canvas_api_key: MappedSetting[Optional[str]] = MappedSetting(default=None)
    panopto_url: MappedSetting[Optional[str]] = MappedSetting(default=None)
    quick_sync_enabled: BoolSetting = BoolSetting(default=False)
    sync_on_start: BoolSetting = BoolSetting(default=False)

    @property
    def canvas_config(self) -> CanvasClientConfig:
        return CanvasClientConfig(
            api_token=self.canvas_api_key, canvas_url=self.canvas_url
        )

    @property
    def panopto_config(self) -> PanoptoClientConfig:
        return PanoptoClientConfig(panopto_url=self.panopto_url)



import logging

from qtpy.QtCore import QByteArray, QSettings

from qcanvas.util.settings._mapped_setting import MappedSetting

_logger = logging.getLogger(__name__)

_default_theme = "auto"


def ensure_theme_is_valid(theme: str) -> str:
    if theme not in ["auto", "light", "dark", "native"]:
        return _default_theme
    else:
        return theme


class ThemeSetting(MappedSetting):
    def __init__(self):
        super().__init__(default=_default_theme)

    def __get__(self, instance, owner):
        return ensure_theme_is_valid(super().__get__(instance, owner))

    def __set__(self, instance, value):
        super().__set__(instance, ensure_theme_is_valid(value))


class _UISettings:
    settings = QSettings("QCanvasTeam", "QCanvas")
    theme: ThemeSetting = ThemeSetting()
    last_geometry: MappedSetting[QByteArray] = MappedSetting()
    last_window_state: MappedSetting[QByteArray] = MappedSetting()



import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import Qt, Signal

from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class _CourseTreeItem(MemoryTreeWidgetItem):
    def __init__(self, course: db.Course, owner: "CourseTree"):
        MemoryTreeWidgetItem.__init__(
            self,
            id=course.id,
            data=course,
            strings=[course.configuration.nickname or course.name],
        )

        self._owner = owner
        self._course = course

        self.setFlags(
            Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEnabled
        )

    def setData(self, column: int, role: int, value: Any):
        if column != 0 or not isinstance(value, str):
            return super().setData(column, role, value)

        value = value.strip()

        if len(value) == 0:
            super().setData(column, role, self._course.name)
            self._owner.course_renamed.emit(self._course, None)
        else:
            super().setData(column, role, value)
            self._owner.course_renamed.emit(self._course, value)


class CourseTree(ContentTree[Sequence[db.Term]]):
    course_renamed = Signal(db.Course, str)

    def __init__(self):
        super().__init__("course_tree", emit_selection_signal_for_type=db.Course)

        self.ui_setup(max_width=250, min_width=150, header_text="Courses")

    def create_tree_items(
        self, terms: List[db.Term], sync_receipt: Optional[SyncReceipt]
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for term in reversed(terms):
            term_widget = self._create_term_widget(term)

            for course in term.courses:
                course_widget = self._create_course_widget(course, sync_receipt)
                # course_widget.renamed.connect(self._on_course_renamed)
                term_widget.addChild(course_widget)

            widgets.append(term_widget)

        return widgets

    def _create_course_widget(
        self, course: db.Course, sync_receipt: Optional[SyncReceipt]
    ) -> _CourseTreeItem:
        course_widget = _CourseTreeItem(course, self)

        is_new = sync_receipt is not None and course.id in sync_receipt.updated_courses

        if is_new:
            self.mark_as_unseen(course_widget)

        return course_widget

    def _create_term_widget(self, term: db.Term) -> MemoryTreeWidgetItem:
        term_widget = MemoryTreeWidgetItem(id=term.id, data=term, strings=[term.name])
        term_widget.setFlags(Qt.ItemFlag.ItemIsEnabled)

        return term_widget

    # @Slot(db.Course, str)
    # def _on_course_renamed(self, course: db.Course, new_name: str) -> None:
    #     self.course_renamed.emit(course, new_name)



import logging
import os
import platform
import sys
from pathlib import Path

import cachetools
import platformdirs
from qtpy.QtCore import QDir, QSettings

_logger = logging.getLogger(__name__)

_is_running_on_windows = platform.system() == "Windows"
_is_running_on_linux = platform.system() == "Linux"
_is_running_as_pyinstaller = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")
_is_running_as_flatpak = os.environ.get("container", "") == "flatpak"


def client_settings() -> QSettings:
    if _is_running_as_pyinstaller and _is_running_on_windows:
        return QSettings(
            str(Path(QDir.homePath()) / ".config" / "QCanvasTeam" / "canvas.ini"),
            QSettings.Format.IniFormat,
        )
    else:
        return QSettings("QCanvasTeam", "QCanvas")


@cachetools.cached(cachetools.LRUCache(maxsize=1))
def root() -> Path:
    root_path = Path()

    if _is_running_as_flatpak:
        root_path = os.environ["XDG_DATA_HOME"]
    elif _is_running_as_pyinstaller:
        root_path = platformdirs.user_data_path("QCanvasReborn", "QCanvasTeam")

    _logger.debug("Root path %s", root_path.absolute())
    print(root_path.absolute())
    return root_path


def ui_storage() -> Path:
    return root() / ".UI"


def data_storage() -> Path:
    return root() / ".DATA"



import logging
from abc import ABCMeta
from pathlib import Path
from threading import Lock

import qcanvas_backend.database.types as db
from qcanvas_backend.database import QCanvasDatabase
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.resources.extracting.extractors import Extractors
from qcanvas_backend.task_master import TaskID
from qtpy.QtCore import QObject, Signal

from qcanvas.backend_connectors.qcanvas_task_master import task_master

_logger = logging.getLogger(__name__)


class _Meta(type(QObject), ABCMeta): ...


class FrontendResourceManager(QObject, ResourceManager, metaclass=_Meta):
    download_finished = Signal(db.Resource)
    download_failed = Signal(db.Resource)

    def __init__(
        self,
        database: QCanvasDatabase,
        download_dest: Path,
        extractors: Extractors = Extractors(),
    ):
        super().__init__(
            database=database, download_dest=download_dest, extractors=extractors
        )
        self._lock = Lock()
        self._download_tasks: dict[str, TaskID] = {}

    # todo will need more signals
    def on_download_progress(
        self, resource: db.Resource, current: int, total: int
    ) -> None:
        with self._lock:
            if resource.id not in self._download_tasks:
                task = TaskID("Download", resource.file_name)
                self._download_tasks[resource.id] = task
            elif current == total and total != 0:
                task = self._download_tasks.pop(resource.id, None)
            else:
                task = self._download_tasks[resource.id]

        task_master.report_progress(task, current, total)

    def on_download_failed(self, resource: db.Resource) -> None:
        with self._lock:
            task = self._download_tasks.pop(
                resource.id, TaskID("Download", resource.file_name)
            )

        task_master.report_failed(task, "Download failed")

        self.download_failed.emit(resource)

    def on_download_finished(self, resource: db.Resource) -> None:
        self.download_finished.emit(resource)



import logging

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager

_logger = logging.getLogger(__name__)


# todo stub for now
class _RM(ResourceManager):
    def on_download_progress(self, resource: db.Resource, current: int, total: int):
        if total == 0 and current == 0:
            _logger.info(f"download of {resource.file_name}: ?%")
        else:
            _logger.info(
                f"download of {resource.file_name}: {(current / total) * 100}%"
            )

    def on_download_failed(self, resource: db.Resource):
        _logger.info(f"download of {resource.file_name} failed")

    def on_download_finished(self, resource: db.Resource):
        _logger.info(f"download of {resource.file_name} finished")
